import kp
import cv2
import os
import numpy as np
import pandas as pd


def softmax(x, axis=1):
    e_x = np.exp(x)
    return e_x / e_x.sum(axis=axis, keepdims=True)


dir_image = "./images"
nef_file = "./models_720.nef"
result_file = "./nef_usb_results.xlsx"
image_files = [f for f in os.listdir(dir_image) if f.endswith(('jpg'))]

# Connect to Kneron device
target_device_type = kp.ProductId.KP_DEVICE_KL720
usb_port_ids = []
device_descriptors = kp.core.scan_devices()
for device_descriptor in device_descriptors.device_descriptor_list:
    if target_device_type == device_descriptor.product_id and device_descriptor.is_connectable:
        usb_port_ids.append(device_descriptor.usb_port_id)

try:
    print("[Connect device]")
    device_group = kp.core.connect_devices(usb_port_ids=[usb_port_ids[0]])
    print('Success')
except Exception as e:
    print(f"Failed to connect devices: {e}")

# Load NEF Model
try:
    print("[Load NEF model]")
    model_nef_descriptor = kp.core.load_model_from_file(device_group=device_group, file_path=nef_file)
    print('Success')
except Exception as e:
    print(f"Failed to load NEF model: {e}")

# Get Model Information on Kneron Device
model_nef_descriptors = []
for usb_port_id in usb_port_ids:
    model_nef_descriptor = kp.core.get_model_info(device_group=device_group, usb_port_id=usb_port_id)
    model_nef_descriptors.append(model_nef_descriptor)

# Process each image
predictions = []
for idx, image_file in enumerate(image_files):
    print(f'Processing image {idx}/{len(image_files)}: {image_file}')
    img = cv2.imread(os.path.join(dir_image, image_file))
    img_rgb565 = cv2.cvtColor(src=img, code=cv2.COLOR_RGB2BGR565)

    generic_raw_image_header = kp.GenericImageInferenceDescriptor(
        model_id=model_nef_descriptor.models[0].id,
        inference_number=0,
        input_node_image_list=[
            kp.GenericInputNodeImage(
                #image=img_bgr565,
                image=img_rgb565,
                image_format=kp.ImageFormat.KP_IMAGE_FORMAT_RGB565,
                resize_mode=kp.ResizeMode.KP_RESIZE_ENABLE,
                padding_mode=kp.PaddingMode.KP_PADDING_DISABLE,
                normalize_mode=kp.NormalizeMode.KP_NORMALIZE_DISABLE
            )
        ]
    )
  
    kp.inference.generic_image_inference_send(device_group=device_group, generic_inference_input_descriptor=generic_raw_image_header)
    generic_raw_result = kp.inference.generic_image_inference_receive(device_group=device_group)
    
    inf_node_output_list = []
    for node_idx in range(generic_raw_result.header.num_output_node):
        inference_float_node_output = kp.inference.generic_inference_retrieve_float_node(
            node_idx=node_idx,
            generic_raw_result=generic_raw_result,
            channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW
        )
        inf_node_output_list.append(inference_float_node_output.ndarray) 
    inf_node_output_array = np.concatenate(inf_node_output_list, axis=0)
    predictions.append(inf_node_output_array)
predictions = np.vstack(predictions).reshape(len(image_files), -1)
predictions = softmax(predictions)
pred_labels = predictions.argmax(axis=1)
pred_probs = predictions.max(axis=1)
print(f"Nef Labels from usb: {pred_labels} \n")
print(f"Nef Probabilities from usb: {pred_probs} \n\n")


df_final_pred = np.vstack([pred_probs, pred_labels]).T
df_final_pred = pd.DataFrame(df_final_pred, columns=["pred prob", "pred label"])
df = pd.DataFrame(image_files, columns=["image name"])
df_final_pred = df.join(df_final_pred, how="outer")
df_full_pred = pd.DataFrame(predictions.T, columns=image_files)
with pd.ExcelWriter(result_file) as writer:  
    df_final_pred.to_excel(writer, sheet_name='final pred')
    df_full_pred.to_excel(writer, sheet_name='full pred')

